;; Generated package description from sxhkdrc-mode.el  -*- no-byte-compile: t -*-
(define-package "sxhkdrc-mode" "0.1.5.0.20230210.82528" "Major mode for sxhkdrc files (Simple X Hot Key Daemon)" '((emacs "27.1")) :commit "60071b73aa2df882fa13febd56437be5a1a9bedf" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou General Issues" . "~protesilaos/general-issues@lists.sr.ht") :url "https://git.sr.ht/~protesilaos/sxhkdrc-mode")
